/*
 * 	hifn_main.c
 * 
 * 2004 Copyright (c) Evgeniy Polyakov <johnpol@2ka.mipt.ru>
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/mod_devicetable.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/mm.h>
#include <linux/highmem.h>
#include <linux/reboot.h>

#include <asm/mach-default/mach_reboot.h>
#include <asm/kmap_types.h>

#include "hifn_main.h"
#include "hifn_reg.h"
#include "hifn_session.h"

#include "../crypto/acrypto.h"
#include "../crypto/crypto_stat.h"
#include "../crypto/crypto_def.h"

#define PCI_VENDOR_ID_HIFN		0x13A3
#define PCI_DEVICE_ID_HIFN_7955		0x0020
#define	PCI_DEVICE_ID_HIFN_7956		0x001d

#undef dprintk

//#define HIFN_TEST
#define HIFN_DEBUG
#ifdef HIFN_DEBUG
#define dprintk(f, a...) 	printk(KERN_INFO f, ##a)
#else
#define dprintk(f, a...)	do {} while (0)
#endif

static struct pci_device_id hifn_pci_tbl[] = 
{
	{ PCI_VENDOR_ID_HIFN, PCI_DEVICE_ID_HIFN_7955, PCI_ANY_ID, PCI_ANY_ID },
	{ PCI_VENDOR_ID_HIFN, PCI_DEVICE_ID_HIFN_7956, PCI_ANY_ID, PCI_ANY_ID },
	{ 0 }
};
MODULE_DEVICE_TABLE(pci, hifn_pci_tbl);

static int hifn_probe(struct pci_dev *, const struct pci_device_id *);
static void hifn_remove(struct pci_dev *);

static struct pci_driver hifn_pci_driver = {
	.name     = "hifn7955",
	.id_table = hifn_pci_tbl,
	.probe    = hifn_probe,
	.remove   = __devexit_p(hifn_remove),
};

static int hifn_dev_number;

u32 hifn_read_0(struct hifn_device *dev, u32 reg)
{
	u32 ret;

	ret = readl((char *)(dev->bar[0]) + reg);

	return ret;
}
u32 hifn_read_1(struct hifn_device *dev, u32 reg)
{
	u32 ret;

	ret = readl((char *)(dev->bar[1]) + reg);

	return ret;
}

void hifn_write_0(struct hifn_device *dev, u32 reg, u32 val)
{
	writel(val, (char *)(dev->bar[0]) + reg);
}
void hifn_write_1(struct hifn_device *dev, u32 reg, u32 val)
{
	writel(val, (char *)(dev->bar[1]) + reg);
}

static void hifn_dump_ring(struct hifn_device *dev)
{
#ifdef HIFN_TEST
	struct hifn_dma *dma = (struct hifn_dma *)dev->desc_virt;
	dma_addr_t dptr = dev->desc_dma;

	dprintk("%4d cmdr: %08x %d\n", 
			dma->cmdi,
			hifn_read_1(dev, HIFN_1_DMA_CRAR), 
			(hifn_read_1(dev, HIFN_1_DMA_CRAR) - 
			 	__cpu_to_le32(dptr + offsetof(struct hifn_dma, cmdr[0])))/sizeof(struct hifn_desc));
	dprintk("%4d srcr: %08x %d\n", 
			dma->srci,
			hifn_read_1(dev, HIFN_1_DMA_SRAR), 
			(hifn_read_1(dev, HIFN_1_DMA_SRAR) - 
			 	__cpu_to_le32(dptr + offsetof(struct hifn_dma, srcr[0])))/sizeof(struct hifn_desc));
	dprintk("%4d dstr: %08x %d\n", 
			dma->dsti,
			hifn_read_1(dev, HIFN_1_DMA_DRAR), 
			(hifn_read_1(dev, HIFN_1_DMA_DRAR) - 
			 	__cpu_to_le32(dptr + offsetof(struct hifn_dma, dstr[0])))/sizeof(struct hifn_desc));
	dprintk("%4d resr: %08x %d\n", 
			dma->resi,
			hifn_read_1(dev, HIFN_1_DMA_RRAR), 
			(hifn_read_1(dev, HIFN_1_DMA_RRAR) - 
			 	__cpu_to_le32(dptr + offsetof(struct hifn_dma, resr[0])))/sizeof(struct hifn_desc));
#endif
}

static void hifn_wait_puc(struct hifn_device *dev)
{
	int i;
	u32 ret;

	for (i=10000; i > 0; --i)
	{
		ret = hifn_read_0(dev, HIFN_0_PUCTRL);
		if (!(ret & HIFN_PUCTRL_RESET))
			break;

		udelay(1);
	}

	if (!i)
		dprintk("%s: Failed to reset PUC unit.\n", dev->name);
}

static void hifn_reset_puc(struct hifn_device *dev)
{
	hifn_write_0(dev, HIFN_0_PUCTRL, HIFN_PUCTRL_DMAENA);
	hifn_wait_puc(dev);
}

static void hifn_stop_device(struct hifn_device *dev)
{
	hifn_write_1(dev, HIFN_1_DMA_CSR,
	    HIFN_DMACSR_D_CTRL_DIS | HIFN_DMACSR_R_CTRL_DIS |
	    HIFN_DMACSR_S_CTRL_DIS | HIFN_DMACSR_C_CTRL_DIS);
	hifn_write_0(dev, HIFN_0_PUIER, 0);
	hifn_write_1(dev, HIFN_1_DMA_IER, 0);
}

static void hifn_reset_dma(struct hifn_device *dev, int full)
{
	hifn_stop_device(dev);
	
	/*
	 * Setting poll frequency and others to 0.
	 */
	hifn_write_1(dev, HIFN_1_DMA_CNFG, HIFN_DMACNFG_MSTRESET | HIFN_DMACNFG_DMARESET | HIFN_DMACNFG_MODE);
	mdelay(1);

	/*
	 * Reset DMA.
	 */
	if (full)
	{
		hifn_write_1(dev, HIFN_1_DMA_CNFG, HIFN_DMACNFG_MODE);
		mdelay(1);
	}
	else
	{
		hifn_write_1(dev, HIFN_1_DMA_CNFG, HIFN_DMACNFG_MODE | HIFN_DMACNFG_MSTRESET);
		hifn_reset_puc(dev);
	}

	hifn_write_1(dev, HIFN_1_DMA_CNFG, HIFN_DMACNFG_MSTRESET | HIFN_DMACNFG_DMARESET | HIFN_DMACNFG_MODE);

	hifn_reset_puc(dev);
}

static u32 hifn_next_signature(u_int32_t a, u_int cnt)
{
	int i;
	u32 v;

	for (i = 0; i < cnt; i++) {

		/* get the parity */
		v = a & 0x80080125;
		v ^= v >> 16;
		v ^= v >> 8;
		v ^= v >> 4;
		v ^= v >> 2;
		v ^= v >> 1;

		a = (v & 1) ^ (a << 1);
	}

	return a;
}

static struct pci2id {
	u_short		pci_vendor;
	u_short		pci_prod;
	char		card_id[13];
} pci2id[] = {
	{
		PCI_VENDOR_ID_HIFN,
		PCI_DEVICE_ID_HIFN_7955,
		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		  0x00, 0x00, 0x00, 0x00, 0x00 }
	}, 
	{
		PCI_VENDOR_ID_HIFN,
		PCI_DEVICE_ID_HIFN_7956,
		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		  0x00, 0x00, 0x00, 0x00, 0x00 }
	}
};

static int hifn_init_pubrng(struct hifn_device *dev)
{
	int i;
	
	hifn_write_1(dev, HIFN_1_PUB_RESET, hifn_read_1(dev, HIFN_1_PUB_RESET) | HIFN_PUBRST_RESET);

	for (i=100; i > 0; --i)
	{
		mdelay(1);

		if ((hifn_read_1(dev, HIFN_1_PUB_RESET) & HIFN_PUBRST_RESET) == 0)
			break;
	}

	if (!i)
		dprintk("Chip %s: Failed to initialise public key engine.\n", dev->name);
	else
	{
		hifn_write_1(dev, HIFN_1_PUB_IEN, HIFN_PUBIEN_DONE);
		dev->dmareg |= HIFN_DMAIER_PUBDONE;
		hifn_write_1(dev, HIFN_1_DMA_IER, dev->dmareg);

		dprintk("Chip %s: Public key engine has been sucessfully initialised.\n", dev->name);
	}

	/*
	 * Enable RNG engine.
	 */

	hifn_write_1(dev, HIFN_1_RNG_CONFIG, hifn_read_1(dev, HIFN_1_RNG_CONFIG) | HIFN_RNGCFG_ENA);
	dprintk("Chip %s: RNG engine has been successfully initialised.\n", dev->name);

	return 0;
}

static int hifn_enable_crypto(struct hifn_device *dev)
{
	u32 dmacfg, addr;
	char *offtbl = NULL;
	int i;
	
	for (i = 0; i < sizeof(pci2id)/sizeof(pci2id[0]); i++) 
	{
		if (pci2id[i].pci_vendor == dev->pdev->vendor && pci2id[i].pci_prod == dev->pdev->device) 
		{
			offtbl = pci2id[i].card_id;
			break;
		}
	}

	if (offtbl == NULL) {
		dprintk("Chip %s: Unknown card!\n", dev->name);
		return -ENODEV;
	}

	dmacfg = hifn_read_1(dev, HIFN_1_DMA_CNFG);

	hifn_write_1(dev, HIFN_1_DMA_CNFG, 
			HIFN_DMACNFG_UNLOCK | HIFN_DMACNFG_MSTRESET | 
			HIFN_DMACNFG_DMARESET | HIFN_DMACNFG_MODE);
	mdelay(1);
	addr = hifn_read_1(dev, HIFN_1_UNLOCK_SECRET1);
	mdelay(1);
	hifn_write_1(dev, HIFN_1_UNLOCK_SECRET2, 0);
	mdelay(1);

	for (i=0; i<12; ++i)
	{
		addr = hifn_next_signature(addr, offtbl[i] + 0x101);
		hifn_write_1(dev, HIFN_1_UNLOCK_SECRET2, addr);

		mdelay(1);
	}
	hifn_write_1(dev, HIFN_1_DMA_CNFG, dmacfg);
	
	dprintk("Chip %s: %s.\n", dev->name, pci_pretty_name(dev->pdev));

	return 0;
}

static void hifn_init_dma(struct hifn_device *dev)
{
	struct hifn_dma *dma = (struct hifn_dma *)dev->desc_virt;
	u32 dptr = dev->desc_dma;
	int i;

	for (i=0; i<HIFN_D_CMD_RSIZE; ++i)
		dma->cmdr[i].p = __cpu_to_le32(dptr + offsetof(struct hifn_dma, command_bufs[i][0]));
	for (i=0; i<HIFN_D_RES_RSIZE; ++i)
		dma->resr[i].p = __cpu_to_le32(dptr + offsetof(struct hifn_dma, result_bufs[i][0]));

	/*
	 * Setup LAST descriptors.
	 */
	dma->cmdr[HIFN_D_CMD_RSIZE].p = __cpu_to_le32(dptr + offsetof(struct hifn_dma, cmdr[0]));
	dma->srcr[HIFN_D_SRC_RSIZE].p = __cpu_to_le32(dptr + offsetof(struct hifn_dma, srcr[0]));
	dma->dstr[HIFN_D_DST_RSIZE].p = __cpu_to_le32(dptr + offsetof(struct hifn_dma, dstr[0]));
	dma->resr[HIFN_D_RES_RSIZE].p = __cpu_to_le32(dptr + offsetof(struct hifn_dma, resr[0]));
	
	dma->cmdu = dma->srcu = dma->dstu = dma->resu = 0;
	dma->cmdi = dma->srci = dma->dsti = dma->resi = 0;
	dma->cmdk = dma->srck = dma->dstk = dma->resk = 0;
}

static void hifn_init_registers(struct hifn_device *dev)
{
	u32 dptr = dev->desc_dma;
	
	/* Initialization magic... */
	hifn_write_0(dev, HIFN_0_PUCTRL, HIFN_PUCTRL_DMAENA);
	hifn_write_0(dev, HIFN_0_FIFOCNFG, HIFN_FIFOCNFG_THRESHOLD);
	hifn_write_0(dev, HIFN_0_PUIER, HIFN_PUIER_DSTOVER);

	dprintk("dma=%08x\n", dptr);

	/* write all 4 ring address registers */
	hifn_write_1(dev, HIFN_1_DMA_CRAR, __cpu_to_le32(dptr + offsetof(struct hifn_dma, cmdr[0])));
	hifn_write_1(dev, HIFN_1_DMA_SRAR, __cpu_to_le32(dptr + offsetof(struct hifn_dma, srcr[0])));
	hifn_write_1(dev, HIFN_1_DMA_DRAR, __cpu_to_le32(dptr + offsetof(struct hifn_dma, dstr[0])));
	hifn_write_1(dev, HIFN_1_DMA_RRAR, __cpu_to_le32(dptr + offsetof(struct hifn_dma, resr[0])));

	//hifn_dump_ring(dev);

	mdelay(2);
#if 0	
	hifn_write_1(dev, HIFN_1_DMA_CSR,
	    HIFN_DMACSR_D_CTRL_DIS | HIFN_DMACSR_R_CTRL_DIS |
	    HIFN_DMACSR_S_CTRL_DIS | HIFN_DMACSR_C_CTRL_DIS |
	    HIFN_DMACSR_D_ABORT | HIFN_DMACSR_D_DONE | HIFN_DMACSR_D_LAST |
	    HIFN_DMACSR_D_WAIT | HIFN_DMACSR_D_OVER |
	    HIFN_DMACSR_R_ABORT | HIFN_DMACSR_R_DONE | HIFN_DMACSR_R_LAST |
	    HIFN_DMACSR_R_WAIT | HIFN_DMACSR_R_OVER |
	    HIFN_DMACSR_S_ABORT | HIFN_DMACSR_S_DONE | HIFN_DMACSR_S_LAST |
	    HIFN_DMACSR_S_WAIT |
	    HIFN_DMACSR_C_ABORT | HIFN_DMACSR_C_DONE | HIFN_DMACSR_C_LAST |
	    HIFN_DMACSR_C_WAIT |
	    HIFN_DMACSR_ENGINE | 
	    HIFN_DMACSR_PUBDONE);
#else
	hifn_write_1(dev, HIFN_1_DMA_CSR,
	    HIFN_DMACSR_C_CTRL_ENA | HIFN_DMACSR_S_CTRL_ENA |
	    HIFN_DMACSR_D_CTRL_ENA | HIFN_DMACSR_R_CTRL_ENA |
	    HIFN_DMACSR_D_ABORT | HIFN_DMACSR_D_DONE | HIFN_DMACSR_D_LAST |
	    HIFN_DMACSR_D_WAIT | HIFN_DMACSR_D_OVER |
	    HIFN_DMACSR_R_ABORT | HIFN_DMACSR_R_DONE | HIFN_DMACSR_R_LAST |
	    HIFN_DMACSR_R_WAIT | HIFN_DMACSR_R_OVER |
	    HIFN_DMACSR_S_ABORT | HIFN_DMACSR_S_DONE | HIFN_DMACSR_S_LAST |
	    HIFN_DMACSR_S_WAIT |
	    HIFN_DMACSR_C_ABORT | HIFN_DMACSR_C_DONE | HIFN_DMACSR_C_LAST |
	    HIFN_DMACSR_C_WAIT |
	    HIFN_DMACSR_ENGINE | 
	    HIFN_DMACSR_PUBDONE);
#endif
	hifn_read_1(dev, HIFN_1_DMA_CSR);

	dev->dmareg |= HIFN_DMAIER_R_DONE | HIFN_DMAIER_C_ABORT |
	    HIFN_DMAIER_D_OVER | HIFN_DMAIER_R_OVER |
	    HIFN_DMAIER_S_ABORT | HIFN_DMAIER_D_ABORT | HIFN_DMAIER_R_ABORT |
	    HIFN_DMAIER_ENGINE;
	dev->dmareg &= ~HIFN_DMAIER_C_WAIT;

	hifn_write_1(dev, HIFN_1_DMA_IER, dev->dmareg);
	hifn_read_1(dev, HIFN_1_DMA_IER);
#if 0	
	hifn_write_0(dev, HIFN_0_PUCNFG, HIFN_PUCNFG_ENCCNFG |
		    HIFN_PUCNFG_DRFR_128 | HIFN_PUCNFG_TCALLPHASES |
		    HIFN_PUCNFG_TCDRVTOTEM | HIFN_PUCNFG_BUS32 |
		    HIFN_PUCNFG_DRAM);
#else
	hifn_write_0(dev, HIFN_0_PUCNFG, 0x10342);
#endif
	hifn_write_1(dev, HIFN_1_PLL, HIFN_PLL_7956);
	
	hifn_write_0(dev, HIFN_0_PUISR, HIFN_PUISR_DSTOVER);
	hifn_write_1(dev, HIFN_1_DMA_CNFG, HIFN_DMACNFG_MSTRESET |
	    HIFN_DMACNFG_DMARESET | HIFN_DMACNFG_MODE | HIFN_DMACNFG_LAST |
	    ((HIFN_POLL_FREQUENCY << 16 ) & HIFN_DMACNFG_POLLFREQ) |
	    ((HIFN_POLL_SCALAR << 8) & HIFN_DMACNFG_POLLINVAL));
}

static void hifn_init_sessions(struct hifn_device *dev)
{
	u32 pucnfg;
	u32 ctxsize = 1;

	pucnfg = hifn_read_0(dev, HIFN_0_PUCNFG);
	
	if (pucnfg & HIFN_PUCNFG_COMPSING) 
	{
		if (pucnfg & HIFN_PUCNFG_ENCCNFG)
			ctxsize = 128;
		else
			ctxsize = 512;
		
		dev->max_sessions = dev->ram_size / ctxsize;
	}
	else
		dev->max_sessions = dev->ram_size / 16384;
	
	dprintk("Chip %s: ram size=%uK, max_sessions=%u.\n", 
			dev->name, dev->ram_size / 1024,
			dev->max_sessions);
}

static int hifn_setup_base_command(struct hifn_device *dev, u8 *buf, u16 dlen, u16 slen, u16 mask, u8 snum)
{
	struct hifn_base_command *base_cmd;
	u8 *buf_pos = buf;
	
	base_cmd = (struct hifn_base_command *)buf_pos;
	base_cmd->masks = __cpu_to_le16(mask);
	base_cmd->total_source_count = __cpu_to_le16(slen & HIFN_BASE_CMD_LENMASK_LO);
	base_cmd->total_dest_count = __cpu_to_le16(dlen & HIFN_BASE_CMD_LENMASK_LO);

	dlen >>= 16;
	slen >>= 16;
	base_cmd->session_num = __cpu_to_le16(snum | 
	    ((slen << HIFN_BASE_CMD_SRCLEN_S) & HIFN_BASE_CMD_SRCLEN_M) |
	    ((dlen << HIFN_BASE_CMD_DSTLEN_S) & HIFN_BASE_CMD_DSTLEN_M));

#ifdef HIFN_TEST	
	dprintk("Chip %s: command bas: stat %8x ier %8x, session_num=%04x\n",
			dev->name, hifn_read_1(dev, HIFN_1_DMA_CSR), hifn_read_1(dev, HIFN_1_DMA_IER),
			__le16_to_cpu(base_cmd->session_num));
#endif
	return sizeof(struct hifn_base_command);
}

static int hifn_setup_crypto_command(struct hifn_device *dev, u8 *buf, u16 dlen, u16 slen, u8 *key, int keylen, u8 *iv, int ivlen, u16 mode)
{
	struct hifn_dma *dma = (struct hifn_dma *)dev->desc_virt;
	struct hifn_crypt_command *cry_cmd;
	u8 *buf_pos = buf;
	u16 cmd_len;

	cry_cmd = (struct hifn_crypt_command *)buf_pos;

	cry_cmd->source_count = __cpu_to_le16(dlen & 0xffff);
	dlen >>= 16;
	cry_cmd->masks = __cpu_to_le16(mode | ((dlen << HIFN_CRYPT_CMD_SRCLEN_S) & HIFN_CRYPT_CMD_SRCLEN_M));
	cry_cmd->header_skip = 0;
	cry_cmd->reserved = 0;

	buf_pos += sizeof(struct hifn_crypt_command);

	dma->cmdu++;
	/*if (dma->cmdu > 1) 
	{
		dev->dmareg |= HIFN_DMAIER_C_WAIT;
		hifn_write_1(dev, HIFN_1_DMA_IER, dev->dmareg);
	}*/

	if (keylen)
	{
		memcpy(buf_pos, key, keylen);
		buf_pos += keylen;
	}
	if (ivlen)
	{
		memcpy(buf_pos, iv, ivlen);
		buf_pos += ivlen;
	}

	cmd_len = buf_pos - buf;

#ifdef HIFN_TEST	
	dprintk("Chip %s: command cry: stat %8x ier %8x\n",
			dev->name, hifn_read_1(dev, HIFN_1_DMA_CSR), hifn_read_1(dev, HIFN_1_DMA_IER));
#endif	
	return cmd_len;
}

static int hifn_setup_src_desc(struct hifn_device *dev, struct scatterlist *sg, int nents)
{
	struct hifn_dma *dma = (struct hifn_dma *)dev->desc_virt;
	int i, mapped_nsg, idx, nbytes;
	u32 last = 0;
	
	mapped_nsg = pci_map_sg(dev->pdev, sg, nents, PCI_DMA_TODEVICE);
	
	idx = dma->srci;

	nbytes = 0;
	for (i=0; i<mapped_nsg; ++i)
	{
		if (i == mapped_nsg - 1)
			last = HIFN_D_LAST;
		
		dma->srcr[idx].p = __cpu_to_le32(sg_dma_address(&sg[i]));
		dma->srcr[idx].l = __cpu_to_le32(sg_dma_len(&sg[i]) | HIFN_D_VALID | 
				HIFN_D_MASKDONEIRQ | HIFN_D_NOINVALID | last);
		
		if (++idx == HIFN_D_SRC_RSIZE) 
		{
			dma->srcr[idx].l = __cpu_to_le32(HIFN_D_VALID | HIFN_D_JUMP | HIFN_D_MASKDONEIRQ | HIFN_D_LAST);
			idx = 0;
		}

		nbytes += sg_dma_len(&sg[i]);
	}
		
	dma->srci = idx;
	dma->srcu += mapped_nsg;
#ifdef HIFN_TEST
	dprintk("Chip %s: command src: stat %8x ier %8x, nbytes=%u.\n",
			dev->name, hifn_read_1(dev, HIFN_1_DMA_CSR), hifn_read_1(dev, HIFN_1_DMA_IER),
			nbytes);
#endif
	return nbytes;
}

static void hifn_setup_res_desc(struct hifn_device *dev)
{
	struct hifn_dma *dma = (struct hifn_dma *)dev->desc_virt;
	
	dma->resr[dma->resi].l = __cpu_to_le32(12 | HIFN_D_VALID | HIFN_D_LAST);
	//dma->resr[dma->resi].l = __cpu_to_le32(HIFN_MAX_RESULT | HIFN_D_VALID | HIFN_D_LAST);
	
	if (++dma->resi == HIFN_D_RES_RSIZE) 
	{
		dma->resr[HIFN_D_RES_RSIZE].l = __cpu_to_le32(HIFN_D_VALID | HIFN_D_JUMP | HIFN_D_MASKDONEIRQ | HIFN_D_LAST);
		dma->resi = 0;
	}
	
	dma->resu++;
}

static void hifn_setup_dst_desc(struct hifn_device *dev, struct scatterlist *sg, int mapped_nsg)
{
	struct hifn_dma *dma = (struct hifn_dma *)dev->desc_virt;
	int idx, i;
	u32 last = 0;
	
	mapped_nsg = pci_map_sg(dev->pdev, sg, mapped_nsg, PCI_DMA_FROMDEVICE);

	idx = dma->dsti;

	for (i=0; i<mapped_nsg; ++i)
	{
		if (i == mapped_nsg - 1)
			last = HIFN_D_LAST;
		
		dma->dstr[idx].p = __cpu_to_le32(sg_dma_address(&sg[i]));
		dma->dstr[idx].l = __cpu_to_le32(sg_dma_len(&sg[i]) | HIFN_D_VALID | 
				HIFN_D_MASKDONEIRQ | HIFN_D_NOINVALID | last);
		
		if (++idx == HIFN_D_DST_RSIZE) 
		{
			dma->dstr[idx].l = __cpu_to_le32(HIFN_D_VALID | HIFN_D_JUMP | 
					HIFN_D_MASKDONEIRQ | HIFN_D_LAST | HIFN_D_NOINVALID);
			idx = 0;
		}
	}
#ifdef HIFN_TEST	
	dprintk("Chip %s: command dst: stat %8x ier %8x\n",
			dev->name, hifn_read_1(dev, HIFN_1_DMA_CSR), hifn_read_1(dev, HIFN_1_DMA_IER));
#endif
	dma->dsti = idx;
	dma->dstu += mapped_nsg;
}

int hifn_setup_session(struct hifn_device *dev, 
		struct scatterlist *src, int src_num,
		struct scatterlist *dst, int dst_num, 
		struct scatterlist *__key, struct scatterlist *__iv, 
		u16 op, u16 type, u16 mode, u8 snum,
		struct crypto_session *s)
{
	struct hifn_dma *dma = (struct hifn_dma *)dev->desc_virt;
	int nbytes, cmd_len, sa_idx;
	u8 *buf, *buf_pos;
	u8 *key, *iv;
	u16 mask;
	u16 dlen, slen;

	dprintk("%s: start: idx=%d.\n", __func__, dma->resi);

	dlen = sg_dma_len(dst);
	slen = sg_dma_len(src);
	sa_idx = dma->resi;

	key = iv = NULL;

	nbytes = hifn_setup_src_desc(dev, src, src_num);

	buf = dma->command_bufs[dma->cmdi];

	buf_pos = buf;

	mask = 0;
	switch (op)
	{
		case CRYPTO_OP_DECRYPT:
			mask = HIFN_BASE_CMD_CRYPT | HIFN_BASE_CMD_DECODE;
			break;
		case CRYPTO_OP_ENCRYPT:
			mask = HIFN_BASE_CMD_CRYPT;
			break;
		case CRYPTO_OP_HMAC:
			mask = HIFN_BASE_CMD_MAC;
			break;
		default:
			dprintk("Unsupported operation %04x.\n", op);
			goto err_out;
	}
	
	buf_pos += hifn_setup_base_command(dev, buf_pos, dlen, slen, mask, snum);
	
	if (op == CRYPTO_OP_ENCRYPT || op == CRYPTO_OP_DECRYPT)
	{
		int keysize, ivsize;
		u16 md;

		keysize = ivsize = 0;
		md = 0;
		
		if (__key)
		{
			key = ((u8 *)page_address(__key->page))+__key->offset;
			keysize = sg_dma_len(__key);
			md |= HIFN_CRYPT_CMD_NEW_KEY;
		}
		if (__iv)
		{
			iv = ((u8 *)page_address(__iv->page))+__iv->offset;
			ivsize = sg_dma_len(__iv);
			md |= HIFN_CRYPT_CMD_NEW_IV;
		}

		switch (mode)
		{
			case CRYPTO_MODE_ECB:
				md |= HIFN_CRYPT_CMD_MODE_ECB;
				break;
			case CRYPTO_MODE_CBC:
				md |= HIFN_CRYPT_CMD_MODE_CBC;
				break;
			case CRYPTO_MODE_CFB:
				md |= HIFN_CRYPT_CMD_MODE_CFB;
				break;
			case CRYPTO_MODE_OFB:
				md |= HIFN_CRYPT_CMD_MODE_OFB;
				break;
			default:
				dprintk("Unsupported mode %04x.\n", mode);
				goto err_out;
		}

		switch (type)
		{
			case CRYPTO_TYPE_AES_128:
				md |= HIFN_CRYPT_CMD_KSZ_128 | HIFN_CRYPT_CMD_ALG_AES;
				break;
			case CRYPTO_TYPE_AES_192:
				md |= HIFN_CRYPT_CMD_KSZ_192 | HIFN_CRYPT_CMD_ALG_AES;
				break;
			case CRYPTO_TYPE_AES_256:
				md |= HIFN_CRYPT_CMD_KSZ_256 | HIFN_CRYPT_CMD_ALG_AES;
				break;
			default:
				dprintk("Unsupported type %04x.\n", type);
				goto err_out;
		}

		buf_pos += hifn_setup_crypto_command(dev, buf_pos, dlen, slen, key, keysize, iv, ivsize, md);
	}

	dprintk("Setting session %d to %p.\n", sa_idx, s);
	dev->sa[sa_idx] = s;
	
	cmd_len = buf_pos - buf;
	//cmd_len = sizeof(struct hifn_base_command) + sizeof(struct hifn_crypt_command) + 16;
	dma->cmdr[dma->cmdi].l = __cpu_to_le32(cmd_len | HIFN_D_VALID | HIFN_D_LAST | HIFN_D_MASKDONEIRQ);

	if (++dma->cmdi == HIFN_D_CMD_RSIZE)
	{
		dma->cmdr[dma->cmdi].l = __cpu_to_le32(HIFN_MAX_COMMAND | HIFN_D_VALID | 
				HIFN_D_LAST | HIFN_D_MASKDONEIRQ | HIFN_D_JUMP);
		dma->cmdi = 0;
	}
	else
		dma->cmdr[dma->cmdi-1].l |= __cpu_to_le32(HIFN_D_VALID);
	
	hifn_setup_dst_desc(dev, dst, dst_num);
	hifn_setup_res_desc(dev);

	mb();

err_out:
	dprintk("%s: finish.\n", __func__);

	return 0;
}

static int hifn_test(struct hifn_device *dev, int encdec, u8 snum)
{
	int n, j, err;
	struct hifn_dma *dma = (struct hifn_dma *)dev->desc_virt;
	u8 key[16], iv[16];
	struct scatterlist k, i;
	
	hifn_dump_ring(dev);

	n = 0;
	
	dprintk("%s: session number=%02x:\n",
			(encdec)?"Encoding":"Decoding",
			snum);
	
	dprintk("%3d src[len=%2d, off=%d]: ", dma->srci, dev->sg[n].length, dev->sg[n].offset);
	for (j=0; j<dev->sg[n].length; ++j)
	{
		dprintk("0x%02x, ", ((unsigned char *)page_address(dev->sg[n].page))[j]);
	}
	dprintk("\n");

	memset(key, 0, sizeof(key));
	memset(iv, 0, sizeof(iv));

	k.page 		= virt_to_page(key);
	k.offset	= offset_in_page(key);
	k.length	= sizeof(key);
	
	i.page 		= virt_to_page(iv);
	i.offset	= offset_in_page(iv);
	i.length	= sizeof(iv);

	err = hifn_setup_session(dev, 
			&dev->sg[n], 1, 
			&dev->sg[n], 1, 
			&k, NULL, 
			(encdec)?CRYPTO_OP_ENCRYPT:CRYPTO_OP_DECRYPT, 
			CRYPTO_TYPE_AES_128, CRYPTO_MODE_ECB,
			snum, NULL);
	if (err)
		return err;

	msleep(200);
	
	hifn_dump_ring(dev);

	dprintk("cmd: i=%d, u=%d, k=%d\n", dma->cmdi, dma->cmdu, dma->cmdk);
	dprintk("src: i=%d, u=%d, k=%d\n", dma->srci, dma->srcu, dma->srck);
	dprintk("dst: i=%d, u=%d, k=%d\n", dma->dsti, dma->dstu, dma->dstk);
	dprintk("res: i=%d, u=%d, k=%d\n", dma->resi, dma->resu, dma->resk);

	return 0;
}

static int hifn_start_device(struct hifn_device *dev)
{
	int err;

	dev->ram_size = 32768;
	
	hifn_reset_dma(dev, 1);
	
	err = hifn_enable_crypto(dev);
	if (err)
		return err;

	hifn_reset_puc(dev);

	hifn_init_dma(dev);

	hifn_init_registers(dev);

	hifn_init_sessions(dev);
	
	hifn_init_pubrng(dev);

	err = hifn_session_init(dev);
	if (err)
		return err;

#if 0
	{
		int i;
		u16 slen, dlen;
		char *ptr;

		dev->sg[0].page = NULL;
		dev->sg[1].page = NULL;
		
		i = 0;
		slen = dlen = 16;
		
		dev->sg[i].page = alloc_pages(GFP_KERNEL, 0);
		if (!dev->sg[i].page)
		{
			dprintk("Chip %s: Failed to allocate test page.\n", dev->name);
			return -ENOMEM;
		}
		dev->sg[i].offset = 0;
		dev->sg[i].length = slen;

		ptr = (char *)page_address(dev->sg[i].page);

		for (i=0; i<slen; ++i)
			ptr[i] = i;
		
		err = 0;
		for (i=0; i<10; ++i)
		{
			int encdec = (i % 2 == 0);
			u8 snum;
			
			snum = i/2;
			err = hifn_test(dev, encdec, snum);
			if (err)
				break;

		}

		__free_pages(dev->sg[0].page, 0);
	}
#endif
	return err;
}

static unsigned long jprev;
static unsigned long intcnt, max_intcnt = 100;

static irqreturn_t hifn_interrupt(int irq, void *data, struct pt_regs *regs)
{
	struct hifn_device *dev = (struct hifn_device *)data;
	struct hifn_dma *dma = (struct hifn_dma *)dev->desc_virt;
	u32 dmacsr, restart;
	int i;

	intcnt++;

	dmacsr = hifn_read_1(dev, HIFN_1_DMA_CSR);
	dprintk("%s: irq: stat=%08x, dmareg=%08x, events=%08x\n",
			dev->name, 
			dmacsr, dev->dmareg,
			dev->dmareg & dmacsr);
	if ((dmacsr & dev->dmareg) == 0)
	{
		/*
		 * Nothing happend.
		 */
		dprintk("Chip %s: Nothing happend in %s.\n",
				dev->name, __func__);

		goto out;
	}

#if 0
	hifn_write_1(dev, HIFN_1_DMA_CSR, dmacsr & dev->dmareg);
#else
	hifn_write_1(dev, HIFN_1_DMA_CSR,
	    HIFN_DMACSR_C_CTRL_ENA | HIFN_DMACSR_S_CTRL_ENA |
	    HIFN_DMACSR_D_CTRL_ENA | HIFN_DMACSR_R_CTRL_ENA |
	    HIFN_DMACSR_D_ABORT | HIFN_DMACSR_D_DONE | HIFN_DMACSR_D_LAST |
	    HIFN_DMACSR_D_WAIT | HIFN_DMACSR_D_OVER |
	    HIFN_DMACSR_R_ABORT | HIFN_DMACSR_R_DONE | HIFN_DMACSR_R_LAST |
	    HIFN_DMACSR_R_WAIT | HIFN_DMACSR_R_OVER |
	    HIFN_DMACSR_S_ABORT | HIFN_DMACSR_S_DONE | HIFN_DMACSR_S_LAST |
	    HIFN_DMACSR_S_WAIT |
	    HIFN_DMACSR_C_ABORT | HIFN_DMACSR_C_DONE | HIFN_DMACSR_C_LAST |
	    HIFN_DMACSR_C_WAIT |
	    HIFN_DMACSR_ENGINE | 
	    HIFN_DMACSR_PUBDONE);
#endif
	dmacsr = hifn_read_1(dev, HIFN_1_DMA_CSR);
	
	if (dmacsr & HIFN_DMACSR_ENGINE)
		hifn_write_0(dev, HIFN_0_PUISR, hifn_read_0(dev, HIFN_0_PUISR));
	if (dmacsr & HIFN_DMACSR_PUBDONE)
		hifn_write_1(dev, HIFN_1_PUB_STATUS, hifn_read_1(dev, HIFN_1_PUB_STATUS) | HIFN_PUBSTS_DONE);
	
	restart = dmacsr & (HIFN_DMACSR_R_OVER | HIFN_DMACSR_D_OVER);
	if (restart)
	{
		dprintk("Chip %s: overrun: dmacsr=%x.\n", dev->name, dmacsr);
	}
	
	restart = dmacsr & (HIFN_DMACSR_C_ABORT | HIFN_DMACSR_S_ABORT | HIFN_DMACSR_D_ABORT | HIFN_DMACSR_R_ABORT);
	if (restart)
	{
		dprintk("Chip %s: abort. Resetting.\n", dev->name);
		hifn_reset_dma(dev, 1);
		hifn_init_dma(dev);
		hifn_init_registers(dev);
	}
	
#if 1
	dmacsr = hifn_read_1(dev, HIFN_1_DMA_CSR);
	dprintk("%s: irq: stat=%08x\n",	dev->name, dmacsr);

	if ((dmacsr & HIFN_DMACSR_C_WAIT)) 
	{
		dprintk("cmd=%x, is waiting.\n", (dmacsr & HIFN_DMACSR_C_WAIT));
		dev->dmareg &= ~(HIFN_DMAIER_C_WAIT | HIFN_DMAIER_D_WAIT);
		hifn_write_1(dev, HIFN_1_DMA_IER, dev->dmareg);
	}
#endif	

	for (i=0; i<HIFN_D_RES_RSIZE; ++i)
	{
		struct hifn_desc *d = &dma->resr[i];
		
		if (!(d->l & __cpu_to_le32(HIFN_D_VALID)))
		{
			if (dev->sa[i])
			{
				hifn_session_complete(dev, i);
			}
		}
		else
			dprintk("i=%d: valid is set.\n", i);
	}

	hifn_dump_ring(dev);

	if (0)
	{
		static int number = 0;
		int i, n;
		unsigned char *buf;

		n = number;
		number++;

		//if (n < 2)
		if (n >= 2)
			n = 1;
		
		n = 0;

		if (dev->sg[n].page)
		{
			pci_unmap_sg(dev->pdev, &dev->sg[n], 1, PCI_DMA_TODEVICE);

			dprintk("%3d dst[len=%u, off=%u]: ", dma->dsti, dev->sg[n].length, dev->sg[n].offset);
			buf = kmap_atomic(dev->sg[n].page, KM_IRQ0) + dev->sg[n].offset;
			for (i=0; i<sg_dma_len(&dev->sg[n]); ++i)
			{
				dprintk("0x%02x, ", buf[i]);
			}
			dprintk("\n");
			kunmap_atomic(buf, KM_IRQ0);
			
			if (dev->sg[n].page && 0)
			{
				__free_pages(dev->sg[n].page, 0);
				dev->sg[n].page = NULL;
			}
		}
	}

out:

	if (time_after(jiffies, jprev) && 0)
	{
		if (intcnt > max_intcnt)
			machine_restart(NULL);
		jprev = jiffies;
		intcnt = 0;
	}

	return IRQ_HANDLED;
}

static int hifn_probe(struct pci_dev *pdev, const struct pci_device_id *id)
{
	int err, i;
	struct hifn_device *dev;
	
	err = pci_enable_device(pdev);
	if (err)
	{
		dprintk("%s: Failed to enable device.\n", pci_name(pdev));
		return err;
	}
	pci_set_master(pdev);

	err = pci_set_dma_mask(pdev, 0xffffffff);
	if (err)
	{
		dprintk("%s: Failed to setup DMA.\n", pci_name(pdev));
		goto err_out_disable_pci_device;
	}

	err = pci_request_regions(pdev, hifn_pci_driver.name);
	if (err)
	{
		dprintk("%s: Failed to request I/O regions.\n", pci_name(pdev));
		goto err_out_disable_pci_device;
	}
	
	if (pci_resource_len(pdev, 0) < HIFN_BAR0_SIZE ||
	    pci_resource_len(pdev, 1) < HIFN_BAR1_SIZE ||
	    pci_resource_len(pdev, 2) < HIFN_BAR2_SIZE)
	{
		dprintk("%s: Broken hardware - I/O regions are too small.\n", pci_name(pdev));
		err = -ENODEV;
		goto err_out_free_regions;
	}

	dev = (struct hifn_device *)kmalloc(sizeof(*dev), GFP_KERNEL);
	if (!dev)
	{
		dprintk("Failed to allocate hifn_device structure.\n");
		err = -ENOMEM;
		goto err_out_free_regions;
	}

	memset(dev, 0, sizeof(dev));

	snprintf(dev->name, sizeof(dev->name), "hifn%d", hifn_dev_number);
	spin_lock_init(&dev->lock);

	pci_set_drvdata(pdev, dev);

	for (i=0; i<3; ++i)
	{
		unsigned long addr, size;

		addr = pci_resource_start(pdev, i);
		size = pci_resource_len(pdev, i);

		dev->bar[i] = ioremap_nocache(addr, size);
		if (!dev->bar[i])
		{
			dprintk("Failed to remap %d PCI bar: addr=0x%lx, size=0x%lx.\n", 
					i, addr, size);
			goto err_out_unmap_bars;
		}
	}

	dev->result_mem = __get_free_pages(GFP_KERNEL, HIFN_MAX_RESULT_ORDER);
	if (!dev->result_mem)
	{
		dprintk("Failed to allocate %d pages for result_mem.\n", HIFN_MAX_RESULT_ORDER);
		goto err_out_unmap_bars;
	}
	memset((void *)dev->result_mem, 0, PAGE_SIZE*(1<<HIFN_MAX_RESULT_ORDER));

	dev->dst = pci_map_single(pdev, (void *)dev->result_mem, PAGE_SIZE << HIFN_MAX_RESULT_ORDER, PCI_DMA_FROMDEVICE);

	dev->desc_virt = pci_alloc_consistent(pdev, sizeof(struct hifn_dma), &dev->desc_dma);
	if (!dev->desc_virt)
	{
		dprintk("Failed to allocate descriptor rings.\n");
		goto err_out_free_result_pages;
	}
	memset(dev->desc_virt, 0, sizeof(struct hifn_dma));

	dev->pdev = pdev;
	dev->irq = pdev->irq;
	err = request_irq(dev->irq, hifn_interrupt, SA_SHIRQ, dev->name, dev);
	if (err)
	{
		dprintk("Failed to request IRQ%d: err=%d.\n", dev->irq, err);
		dev->irq = 0;
		goto err_out_free_desc;
	}

	hifn_dev_number++;

	dprintk("HIFN crypto accelerator card at %s has been successfully registered as %s: id=%08x.\n",
			pci_name(pdev), dev->name, hifn_read_1(dev, HIFN_CHIP_ID));

	err = hifn_start_device(dev);
	if (err)
		goto err_out_free_irq;
	
	return 0;

err_out_free_irq:
	free_irq(pdev->irq, dev);

err_out_free_desc:
	pci_free_consistent(pdev, sizeof(struct hifn_dma), dev->desc_virt, dev->desc_dma);
	
err_out_free_result_pages:
	pci_unmap_single(pdev, dev->dst, PAGE_SIZE << HIFN_MAX_RESULT_ORDER, PCI_DMA_FROMDEVICE);
	free_pages(dev->result_mem, HIFN_MAX_RESULT_ORDER);
	
err_out_unmap_bars:
	for (i=0; i<3; ++i)
		if (dev->bar[i])
			iounmap(dev->bar[i]);
	
err_out_free_regions:
	pci_release_regions(pdev);
	
err_out_disable_pci_device:
	pci_disable_device(pdev);

	return err;
}

static void hifn_remove(struct pci_dev *pdev)
{
	int i;
	struct hifn_device *dev;
	
	dev = pci_get_drvdata(pdev);

	if (dev)
	{
		hifn_session_fini(dev);
		hifn_reset_dma(dev, 1);
		hifn_stop_device(dev);

		if (dev->irq)
			free_irq(dev->irq, dev->name);
		pci_free_consistent(pdev, sizeof(struct hifn_dma), dev->desc_virt, dev->desc_dma);
		pci_unmap_single(pdev, dev->dst, PAGE_SIZE << HIFN_MAX_RESULT_ORDER, PCI_DMA_FROMDEVICE);
		free_pages(dev->result_mem, HIFN_MAX_RESULT_ORDER);
		for (i=0; i<3; ++i)
			if (dev->bar[i])
				iounmap(dev->bar[i]);

		kfree(dev);
	}
	
	pci_release_regions(pdev);
	pci_disable_device(pdev);
}

static int __devinit hifn_init(void)
{
	int err;
	
	err = pci_register_driver(&hifn_pci_driver);
	if (err < 0)
	{
		dprintk("Failed to register PCI driver for %s device.\n", hifn_pci_driver.name);
		return -ENODEV;
	}

	printk(KERN_INFO "Driver for HIFN 795x crypto accelerator chip has been successfully registered.\n");
	
	return 0;
}

static void __devexit hifn_fini(void)
{
	pci_unregister_driver(&hifn_pci_driver);
	
	printk(KERN_INFO "Driver for HIFN 795x crypto accelerator chip has been successfully deregistered.\n");
}

module_init(hifn_init);
module_exit(hifn_fini);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Evgeniy Polyakov <johnpol@2ka.mipt.ru>");
MODULE_DESCRIPTION("Driver for HIFN 7955/7956 crypto accelerator chip.");
