/*
 * 	hifn_session.c
 * 
 * 2004 Copyright (c) Evgeniy Polyakov <johnpol@2ka.mipt.ru>
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>

#include "../crypto/acrypto.h"
#include "../crypto/crypto_stat.h"
#include "../crypto/crypto_def.h"
#include "../crypto/crypto_route.h"

#include "hifn_main.h"
#include "hifn_reg.h"
#include "hifn_session.h"

static struct crypto_capability hifn_caps[] = 	{
		{CRYPTO_OP_ENCRYPT, CRYPTO_TYPE_AES_128, CRYPTO_MODE_ECB, HIFN_D_SRC_RSIZE},
		{CRYPTO_OP_DECRYPT, CRYPTO_TYPE_AES_128, CRYPTO_MODE_ECB, HIFN_D_SRC_RSIZE},
		{CRYPTO_OP_ENCRYPT, CRYPTO_TYPE_AES_128, CRYPTO_MODE_CBC, HIFN_D_SRC_RSIZE},
		{CRYPTO_OP_DECRYPT, CRYPTO_TYPE_AES_128, CRYPTO_MODE_CBC, HIFN_D_SRC_RSIZE},
		{CRYPTO_OP_ENCRYPT, CRYPTO_TYPE_AES_128, CRYPTO_MODE_CFB, HIFN_D_SRC_RSIZE},
		{CRYPTO_OP_DECRYPT, CRYPTO_TYPE_AES_128, CRYPTO_MODE_CFB, HIFN_D_SRC_RSIZE},
		{CRYPTO_OP_ENCRYPT, CRYPTO_TYPE_AES_128, CRYPTO_MODE_OFB, HIFN_D_SRC_RSIZE},
		{CRYPTO_OP_DECRYPT, CRYPTO_TYPE_AES_128, CRYPTO_MODE_OFB, HIFN_D_SRC_RSIZE},
		
		{CRYPTO_OP_ENCRYPT, CRYPTO_TYPE_AES_192, CRYPTO_MODE_ECB, HIFN_D_SRC_RSIZE},
		{CRYPTO_OP_DECRYPT, CRYPTO_TYPE_AES_192, CRYPTO_MODE_ECB, HIFN_D_SRC_RSIZE},
		{CRYPTO_OP_ENCRYPT, CRYPTO_TYPE_AES_192, CRYPTO_MODE_CBC, HIFN_D_SRC_RSIZE},
		{CRYPTO_OP_DECRYPT, CRYPTO_TYPE_AES_192, CRYPTO_MODE_CBC, HIFN_D_SRC_RSIZE},
		{CRYPTO_OP_ENCRYPT, CRYPTO_TYPE_AES_192, CRYPTO_MODE_CFB, HIFN_D_SRC_RSIZE},
		{CRYPTO_OP_DECRYPT, CRYPTO_TYPE_AES_192, CRYPTO_MODE_CFB, HIFN_D_SRC_RSIZE},
		{CRYPTO_OP_ENCRYPT, CRYPTO_TYPE_AES_192, CRYPTO_MODE_OFB, HIFN_D_SRC_RSIZE},
		{CRYPTO_OP_DECRYPT, CRYPTO_TYPE_AES_192, CRYPTO_MODE_OFB, HIFN_D_SRC_RSIZE},
		
		{CRYPTO_OP_ENCRYPT, CRYPTO_TYPE_AES_256, CRYPTO_MODE_ECB, HIFN_D_SRC_RSIZE},
		{CRYPTO_OP_DECRYPT, CRYPTO_TYPE_AES_256, CRYPTO_MODE_ECB, HIFN_D_SRC_RSIZE},
		{CRYPTO_OP_ENCRYPT, CRYPTO_TYPE_AES_256, CRYPTO_MODE_CBC, HIFN_D_SRC_RSIZE},
		{CRYPTO_OP_DECRYPT, CRYPTO_TYPE_AES_256, CRYPTO_MODE_CBC, HIFN_D_SRC_RSIZE},
		{CRYPTO_OP_ENCRYPT, CRYPTO_TYPE_AES_256, CRYPTO_MODE_CFB, HIFN_D_SRC_RSIZE},
		{CRYPTO_OP_DECRYPT, CRYPTO_TYPE_AES_256, CRYPTO_MODE_CFB, HIFN_D_SRC_RSIZE},
		{CRYPTO_OP_ENCRYPT, CRYPTO_TYPE_AES_256, CRYPTO_MODE_OFB, HIFN_D_SRC_RSIZE},
		{CRYPTO_OP_DECRYPT, CRYPTO_TYPE_AES_256, CRYPTO_MODE_OFB, HIFN_D_SRC_RSIZE},
};
static int hifn_data_ready_reentry;

static int hifn_session_process(struct crypto_session *s)
{
	int err;
	struct crypto_device *cdev;
	struct hifn_device *dev;
	struct crypto_data *d = &s->data;

	cdev = crypto_route_get_current_device(s);
	if (!cdev)
		return -ENODEV;

	dev = cdev->priv;

	err = hifn_setup_session(dev, 
			&d->sg_src, d->sg_src_num,
			&d->sg_dst, d->sg_src_num,
			&d->sg_key, (s->ci.mode != CRYPTO_MODE_ECB)?&d->sg_iv:NULL, 
			s->ci.operation, s->ci.type, s->ci.mode,
			s->ci.id & 0xff, s);

	crypto_device_put(cdev);
	return err;
}

void hifn_session_complete(struct hifn_device *dev, int snum)
{
	struct crypto_session *s, *n;
	struct crypto_device *cdev = dev->priv;

	dprintk("%s: start.\n", __func__);

	spin_lock_irq(&dev->lock);

	if (!dev->sa[snum])
		goto err_out_no_session;

	dprintk("Found session %d %p\n", snum, dev->sa[snum]);

	list_for_each_entry_safe(s, n, &cdev->session_list, dev_queue_entry)
	{
		if (dev->sa[snum] != s)
			continue;

		spin_lock(&s->lock);
		if (!session_completed(s) && session_is_processed(s))
		{
			dprintk("Completing session %d - %llu [%llu].\n", 
					snum, s->ci.id, s->ci.dev_id);
			crypto_stat_complete_inc(s);
			crypto_session_dequeue_route(s);
			complete_session(s);
			stop_process_session(s);

			dev->sa[snum] = NULL;

			crypto_device_put(dev->priv);

			spin_unlock(&s->lock);
			break;
		}
		spin_unlock(&s->lock);
	}

err_out_no_session:
	spin_unlock_irq(&dev->lock);

	dprintk("%s: finish.\n", __func__);
}

static void hifn_data_ready(struct crypto_device *dev)
{
	struct crypto_session *s, *n;
	int err;

	if (hifn_data_ready_reentry)
		return;

	hifn_data_ready_reentry++;

	list_for_each_entry_safe(s, n, &dev->session_list, dev_queue_entry)
	{
		if (!session_completed(s) && !session_is_processed(s))
		{
			start_process_session(s);
			err = hifn_session_process(s);
			if (err)
			{
				crypto_session_dequeue_route(s);
				broke_session(s);
				complete_session(s);
				stop_process_session(s);
			}
		}
	}
	
	hifn_data_ready_reentry--;
}

int hifn_session_init(struct hifn_device *dev)
{
	int err;
	struct crypto_device *cdev;

	cdev = kmalloc(sizeof(*cdev), GFP_KERNEL);
	if (!cdev)
	{
		dprintk("Failed to allocate new crypto_device structure.\n");
		return -ENOMEM;
	}

	memset(cdev, 0, sizeof(*cdev));

	cdev->cap 		= hifn_caps;
	cdev->cap_number	= sizeof(hifn_caps)/sizeof(hifn_caps[0]);
	cdev->priv		= dev;
	cdev->data_ready	= &hifn_data_ready;
	snprintf(cdev->name, sizeof(cdev->name), "%s", dev->name);

	dev->priv		= cdev;

	memset(dev->sa, 0, sizeof(dev->sa));
	
	err = crypto_device_add(cdev);
	if (err)
	{
		kfree(cdev);
		return err;
	}

	dprintk(KERN_INFO "%s crypto device driver is loaded.\n", dev->name);

	return err;
}

void hifn_session_fini(struct hifn_device *dev)
{
	struct crypto_device *cdev = dev->priv;
	
	crypto_device_remove(cdev);
	
	dprintk(KERN_INFO "%s crypto device driver is unloaded.\n", dev->name);
}
